/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */
 ///
_PUBLIC_
// Constructor
  undo::undo (ptGraphicTrees pGA, ptGraphicTablesManager pTab, ptCurrentObjectsManager pLoc)
{
  the_graphicalTrees = pGA;
  the_graphicTablesManager = pTab;
  pCurObjManager = pLoc;
}


/**
  Fonction qui initialise la structure info_undo. Elle est
  appelee par le superviseur des qu'on descend ou remonte dans la
  hierarchie des arbres et dès que l'on fait une nouvelle action
  (déplacement, justification, changement de taille, encapsulation...)
*/
void
undo::init_undo ()
{
  info_undo.typ_obj_recup = typ_vide;
  info_undo.action = C_not_impl;
  info_undo.niveau_recup = info_undo.recup_rg = 0;
  info_undo.obj_recup = info_undo.oper_recup = 0;
  SET_NILTREE (info_undo.arb_recup);
  SET_NILTREE (info_undo.p_recup);
  SET_NILTREE (info_undo.attr_recup);
  SET_NILTREE (info_undo.arb_recup_inte);
}


/**
  Fonction appelée lorsqu'on rentre du texte Signal (après avoir
  cliqué sur le bouton "confirm") ou lorsqu'on fait un "paste"
  sur un port (pour remplacer sa déclaration par celle d'un autre).
  ?p_sig : arbre Signal à sauvegarder
  ?arb_attr : booléen à vrai si on sauvegarde l'attribut graphique
              vaut vrai pour du texte contenu dans des btes, faux
	      pour les ports.
  ?no_obj : numéro du port qu'on sauvegarde, vaut 0 si on sauvegarde
            du texte associé à une boite.
*/
void
undo::get_undo_text (ntrees p_sig, boolean arb_attr, int no_obj)
{
  objet_type t;

  info_undo.action = C_text_win;
  info_undo.obj_recup = no_obj;
  if (IS_NILTREE (p_sig))
    {
      SET_NILTREE (info_undo.arb_recup);
      info_undo.oper_recup = opldecs;
      info_undo.typ_obj_recup = typ_text;
    }
  else
    {
      if (STree_isNameOfModel (p_sig))
	t = typ_text_nom_proc;
      else if (STree_isParametersOfModel (p_sig))
	t = typ_text_param_proc;
      else
	t = typ_text;
      info_undo.typ_obj_recup = t;
      info_undo.oper_recup = noper (p_sig);
      if (arb_attr)
	{
	  info_undo.recup_rg = nposit (p_sig);
	  info_undo.arb_recup = the_graphicalTrees->sauver_arbre (p_sig, pk_true);
	  info_undo.attr_recup = the_graphicalTrees->sauver_arbre (nattrntrees (p_sig, GRAPHI_ATTRNUMBER), pk_false);
	}
      else
	info_undo.arb_recup = the_graphicalTrees->sauver_arbre (p_sig, pk_true);
    }
}


/**
  Fonction appelée lorsqu'on modifie les déclarations locales d'un modèle
  (constantes, types, signaux). Elle construit une nouvelle liste des 
  déclarations a partir des signaux locaux, constantes et types contenus 
  dans ?p_loc 
  ?p_loc :  liste des declarations locales d'un modèle (peut contenir des
  sous-modèles).
*/
void
undo::get_undo_sigloc (ntrees p_loc)
{
  ntrees list_sig, fils;
  int nb_elem, j, i;

  list_sig = LL_make_ldecl_emptyList ();
  nb_elem = ntreearity (p_loc);
  j = 1;
  for (i = 1; i <= nb_elem; i++)
    {
      fils = nchild (j, p_loc);
      if (!STree_isModel (fils))
	{
	  fils = ndestroyListElem (p_loc, j);	/* on enleve les declarations de signaux */
	  ntreeList_appendLast (list_sig, fils);	/* et on les insere dans la nouvelle liste */
	}
      else
	j++;
    }
  info_undo.typ_obj_recup = typ_text;
  info_undo.oper_recup = opldecl;
  info_undo.arb_recup = list_sig;
  info_undo.action = C_text_win;
}


/**
  Fonction appelée lorsqu'on modifie l'interface d'une boite.
  ?p_gra : arbre représentant la boite avec ses ports.
  Si ?p_inte est non null, alors on est au niveau d'un modèle
  et ?p_inte représente l'interface du modèle, on le sauvegarde 
  aussi.
*/
void
undo::get_undo_profil (ntrees p_gra, ntrees p_inte)
{
  // printf("DEBUG entering undo::get_undo_profil\n"); fflush(stdout);
  info_undo.oper_recup = opinte;
  info_undo.arb_recup = nattrcarrier (p_gra);
  info_undo.action = C_text_win;
  info_undo.attr_recup = the_graphicalTrees->sauver_arbre (p_gra, pk_false);
  if (ISNOT_NILTREE (p_inte))
    info_undo.arb_recup_inte = the_graphicalTrees->sauver_arbre (p_inte, pk_false);
  info_undo.typ_obj_recup = typ_text;
  // printf("DEBUG  exiting undo::get_undo_profil\n"); fflush(stdout);
}


/**
  Fonction appelee par les fonctions : justifier_ports, detruire_obj, 
  taille_bte, encapsuler_bte, decapsuler_bte et depl_obj
  ?bte : numéro de la boite à sauvegarder
  ?force : a vrai pour encapsuler_bte et decapsuler_bte, à faux sinon.
*/
void
undo::get_undo_bte (int action, int bte, boolean force)
{
  ntrees arb_gra, p_cour;

  // printf("DEBUG entering undo::get_undo_bte\n"); fflush(stdout);
  info_undo.typ_obj_recup = typ_bte;
  info_undo.action = action;
  p_cour = pCurObjManager->get_arb_cou_engl ();
  arb_gra = the_graphicTablesManager->graphObj_get_tree (bte);
  if (STree_isLocalDeclList (p_cour))
    {
      if ((info_undo.action == C_cut) || (info_undo.action == C_cut_graph))
	{
	  info_undo.p_recup = the_graphicalTrees->get_arb_sig (&arb_gra);
	  info_undo.recup_rg = nposit (info_undo.p_recup);
	  info_undo.arb_recup = the_graphicalTrees->sauver_arbre (info_undo.p_recup, pk_true);
	}
      else
	{
	  info_undo.p_recup = arb_gra;
	  info_undo.recup_rg = nposit (nattrcarrier (arb_gra));
	  info_undo.arb_recup = the_graphicalTrees->sauver_arbre (arb_gra, pk_false);
	}
      info_undo.niveau_recup = 1;
    }
  else if (STree_isModel (p_cour))
    {
      info_undo.p_recup = STreeModel_getBody (p_cour);
      info_undo.arb_recup = the_graphicalTrees->sauver_arbre (info_undo.p_recup, pk_false);
      info_undo.niveau_recup = 0;
      info_undo.recup_rg = 0;
    }
  else
    {
      info_undo.p_recup = p_cour;
      info_undo.arb_recup = the_graphicalTrees->sauver_arbre (p_cour, pk_false);
      if (STree_isModel (p_cour) || STree_isBodyTree (p_cour))
	info_undo.niveau_recup = 0;
      else
	info_undo.niveau_recup = 1;
      info_undo.recup_rg = 0;
    }
  // printf("DEBUG  exiting undo::get_undo_bte\n"); fflush(stdout);
}


/**
  Sauvegarde en prévision de problème lors de la compilation d'une architecture
  ?p_archi est le modèle de processus qui va etre compilé avec "compil archi"
  Il contient la description de l'architecture.
*/
void
undo::get_undo_archi (ntrees p_archi)
{
  info_undo.typ_obj_recup = typ_bte;
  info_undo.oper_recup = opprocess;
  info_undo.arb_recup = the_graphicalTrees->sauver_arbre (p_archi, pk_false);
  info_undo.p_recup = nfather (p_archi);
  info_undo.recup_rg = nposit (p_archi);
  info_undo.niveau_recup = 1;
}


 /**
   Fonction appelée pour faire le undo sur une destruction, un déplacement, 
   un changement de taille, encapsulation ou decapsulation.
 */
void
undo::set_undo_bte ()
{
  ntrees arb_sig, p_save, arb_cou;
  int bte;
  typcoord halo[2];

  // printf("DEBUG entering undo::set_undo_bte\n"); fflush(stdout);
  arb_cou = pCurObjManager->get_arb_cou_engl ();
  p_save = info_undo.arb_recup;
  if (STree_isLocalDeclList (arb_cou) && (info_undo.action != C_cut) && (info_undo.action != C_cut_graph))
    {
      bte = the_graphicalTrees->tree_get_graphObj (p_save);
      the_graphicTablesManager->enlever_obj (bte);
      the_graphicalTrees->calcul_halo (p_save, halo);
      the_graphicTablesManager->inser_obj (bte, p_save, halo);
      arb_sig = nattrcarrier (info_undo.p_recup);
      nmakeattrntrees (arb_sig, GRAPHI_ATTRNUMBER, p_save);
    }
  else
    {
      the_graphicalTrees->reconstruire_tab (p_save, pk_true, pk_false, 0, 0);
      if (!STree_isModel (arb_cou))
	{
	  if ((info_undo.action != C_cut) && (info_undo.action != C_cut_graph))
	    STree_replaceSon (info_undo.p_recup, p_save, 0);
	  if (info_undo.p_recup == arb_cou)
	    {
	      if ((info_undo.action == C_cut) || (info_undo.action == C_cut_graph))
		STree_replaceSon (info_undo.p_recup, p_save, 0);
	      pCurObjManager->set_arb_cou_engl (p_save);
	      pCurObjManager->set_bte_cou_engl (the_graphicalTrees->get_arb_boite (p_save));
	    }
	  else
	    {
	      if ((info_undo.action == C_cut) || (info_undo.action == C_cut_graph))
		ninsertright (info_undo.recup_rg - 1, p_save, arb_cou);	// March 2007: insertTree(arb_cou, p_save, info_undo.recup_rg);
	    }
	}
      else
	STree_replaceSon (info_undo.p_recup, p_save, 0);
      if (((info_undo.action == C_cut) || (info_undo.action == C_cut_graph)) && STree_isModel (p_save))
	the_graphicalTrees->create_modelInJTree (p_save, pk_false, pk_true);	/* affichage hierarchie processus */
    }
  // printf("DEBUG  exiting undo::set_undo_bte\n"); fflush(stdout);
}


/**
  Fonction appelée si la compilation de l'architecture décrite par
  le modèle ?p_archi s'est mal passée. On remplace alors ?p_archi 
  par ce qu'il était avant la compil.
*/
void
undo::set_undo_archi (ntrees p_archi)
{
  int n;
  ntrees p_save, pere;

  if (ISNOT_NILTREE (p_archi))
    {
      p_save = info_undo.arb_recup;
      the_graphicalTrees->reconstruire_tab (p_save, pk_true, pk_false, 0, 0);
      pere = info_undo.p_recup;
      n = ntreearity (pere);
      STree_replaceSon (pere, p_save, info_undo.recup_rg);
      the_graphicalTrees->replaceInHierar (p_save);
      pCurObjManager->set_arb_cou_engl (p_save);
      pCurObjManager->set_bte_cou_engl (the_graphicalTrees->get_arb_boite (p_save));
    }
}

/**
  Fonction appelee lorsque l'utilisateur clique sur le bouton "undo".
*/
void
undo::applyUndo (void)
{
  ntrees arb_port;

  switch (info_undo.typ_obj_recup)
    {
    case typ_bte:
    case typ_port:
    case typ_seg:
      set_undo_bte ();
      break;
    case typ_text_param_proc:
    case typ_text_nom_proc:
      set_undo_text ();
      break;
    case typ_text:
      switch (info_undo.oper_recup)
	{
	case opdecsig:
	case opldecs:
	case opnil:
	  arb_port = the_graphicTablesManager->graphObj_get_tree (info_undo.obj_recup);
	  if (IS_NILTREE (info_undo.arb_recup))
	    ndestroyattr (arb_port, SIGNAL_ATTRNUMBER);
	  else
	    the_graphicalTrees->assoc_decl_port (arb_port, info_undo.arb_recup);
	  break;
	case opldecl:
	  STree_replaceLocalDecl (pCurObjManager->getModelInWhich_cou_engl (), info_undo.arb_recup);
	  break;
	case opinte:
	  set_undo_profil ();
	  break;
	case opprocess:
	case opfunc:
	  set_undo_proc ();
	  break;
	default:
	  set_undo_text ();
	}
    default:
      ;
    }
  init_undo ();
}
